﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Controls.Primitives;

namespace DiagramLibrary.Adorner
{
    public partial class Adorner : UserControl
    {
        private double _x = 100;
        private double _y = 0;
        private double _angle = 0;
        private Dictionary<Thumb, DragData> _dragData = new Dictionary<Thumb, DragData>();
        private const double PI = 3.14159265358979323846264338327950288;

        private IAdornedObject _adornedObject = null;

        /// <summary>
        /// Sets the object that is displayed within this adorner
        /// The adorner will move around the object set
        /// </summary>
        /// <param name="notify">the object being adorned</param>
        public void SetAdornedObject(IAdornedObject notify)
        {
            _adornedObject = notify;
            _x = _adornedObject.X;
            _y = _adornedObject.Y;
            _angle = _adornedObject.Angle;

            Width = _adornedObject.Width;
            Height = _adornedObject.Height;
            trTranslate.X = _x;
            trTranslate.Y = _y;
            trRotate.Angle = _angle;
        }

        #region DragData

        /// <summary>
        /// Contains dragging information for a specified adorner
        /// </summary>
        private class DragData
        {
            /// <summary>
            /// Angle of the cursor based on the adorner location
            /// </summary>
            public double CursorAngle;

            /// <summary>
            /// Type of drag being performed
            /// </summary>
            public DragType DragType;

            /// <summary>
            /// true if rotating from one of the left adorners, false otherwise
            /// the (RotateIsLeft, RotateIsTop) combination is used to determine angle between the edge where the adorner is at and the center
            /// </summary>
            public bool RotateIsLeft = true;

            /// <summary>
            /// true if rotating from one of the top adorners, false otherwise
            /// the (RotateIsLeft, RotateIsTop) combination is used to determine angle between the edge where the adorner is at and the center
            /// </summary>
            public bool RotateIsTop = true;

            /// <summary>
            /// Calculated on start of rotation, this is the angle between the currently dragged adorner and the center
            /// used to "compensate" the angle between the mouse and center to ensure all 3 points: adorner, mouse, center 
            /// are on the same line
            /// </summary>
            public double RotateAdornerAngleRad = 0;

            /// <summary>
            /// Used for resize adorners only.
            /// Shows how much the width will change when movin mouse one positive unit.
            /// typically -1, 0, or 1
            /// </summary>
            public double WidthChange = 0;

            /// <summary>
            /// Used for resize adorners only.
            /// Shows how much the height will change when movin mouse one positive unit.
            /// typically -1, 0, or 1 
            /// </summary>
            public double HeightChange = 0;

            /// <summary>
            /// Selects left or right edge to not move during resize
            /// -1: right
            /// 1: left
            /// </summary>
            public double HSignInverse = 1;

            /// <summary>
            /// Selects top or bottom edge to not move during resize
            /// -1: top
            /// 1: bottom
            /// </summary>
            public double VSign = 0;
        };

        #endregion

        #region DragType

        private enum DragType
        {
            None,
            Size,
            Rotate,
            Move,
            Connector
        };

        #endregion

        private DragType _currentDragOperation = DragType.None;

        #region ctor

        public Adorner()
        {
            InitializeComponent();

            _dragData.Add(resize00, new DragData() { CursorAngle = 45, DragType = DragType.Size, WidthChange = -1, HeightChange = -1, HSignInverse = -1, VSign = 1 });
            _dragData.Add(resize10, new DragData() { CursorAngle = 135, DragType = DragType.Size, WidthChange = 1, HeightChange = -1, HSignInverse = 1, VSign = 1 });
            _dragData.Add(resize01, new DragData() { CursorAngle = 135, DragType = DragType.Size, WidthChange = -1, HeightChange = 1, HSignInverse = -1, VSign = -1 });
            _dragData.Add(resize11, new DragData() { CursorAngle = 45, DragType = DragType.Size, WidthChange = 1, HeightChange = 1, HSignInverse = 1, VSign = -1 });

            _dragData.Add(resizeTop, new DragData() { CursorAngle = 90, DragType = DragType.Size, WidthChange = 0, HeightChange = -1, HSignInverse = -1, VSign = 1 });
            _dragData.Add(resizeBottom, new DragData() { CursorAngle = 90, DragType = DragType.Size, WidthChange = 0, HeightChange = 1, HSignInverse = 1, VSign = -1 });
            _dragData.Add(resizeLeft, new DragData() { CursorAngle = 0, DragType = DragType.Size, WidthChange = -1, HeightChange = 0, HSignInverse = -1, VSign = 1 });
            _dragData.Add(resizeRight, new DragData() { CursorAngle = 0, DragType = DragType.Size, WidthChange = 1, HeightChange = 0, HSignInverse = 1, VSign = -1 });

            _dragData.Add(rotate00, new DragData() { CursorAngle = 0, DragType = DragType.Rotate, RotateIsLeft = true, RotateIsTop = true });
            _dragData.Add(rotate10, new DragData() { CursorAngle = 90, DragType = DragType.Rotate, RotateIsLeft = false, RotateIsTop = true });
            _dragData.Add(rotate01, new DragData() { CursorAngle = 270, DragType = DragType.Rotate, RotateIsLeft = true, RotateIsTop = false });
            _dragData.Add(rotate11, new DragData() { CursorAngle = 180, DragType = DragType.Rotate, RotateIsLeft = false, RotateIsTop = false });

            //_dragData.Add(connectorTop, new DragData() { CursorAngle = 90, DragType = DragType.Size, WidthChange = 0, HeightChange = -1, HSignInverse = -1, VSign = 1 });
            //_dragData.Add(connectorBottom, new DragData() { CursorAngle = 90, DragType = DragType.Size, WidthChange = 0, HeightChange = 1, HSignInverse = 1, VSign = -1 });

            _dragData.Add(thumbMove, new DragData() { CursorAngle = 0, DragType = DragType.Move });
        }

        #endregion

        #region Rotate

        Point _mousePosRelativeToThis;
        private void rotate_DragStarted(object sender, System.Windows.Controls.Primitives.DragStartedEventArgs e)
        {
            _currentDragOperation = DragType.Rotate;
            _draggedThumb = (Thumb)sender;
            Point result = trRotate.Inverse.Transform(new Point(e.HorizontalOffset - _x, e.VerticalOffset - _y));
            _mousePosRelativeToThis.X = result.X;
            _mousePosRelativeToThis.Y = result.Y;
            DragData data = _dragData[_draggedThumb];
            data.RotateAdornerAngleRad = GetRotateAdornerAngle(data, Width / 2, Height / 2);
        }

        #endregion

        #region Dragging and Cursors

        private void thumb_MouseEnter(object sender, MouseEventArgs e)
        {
            Thumb thumb = (Thumb)sender;
            DragData data = _dragData[thumb];
            ShowCursor(data.DragType, e.GetPosition(this), data.CursorAngle);
        }

        private void thumb_MouseLeave(object sender, MouseEventArgs e)
        {
            if (_draggedThumb == null)
            {
                HideAllCursors();
            }
        }

        Point _lastMousePos;
        bool _lastMousePosInitialized = false;

        /// <summary>
        /// Retrieves the angle between the edge for a given rotate adorner and the center
        /// </summary>
        /// <param name="data"></param>
        /// <param name="centerX"></param>
        /// <param name="centerY"></param>
        /// <returns></returns>
        private double GetRotateAdornerAngle(DragData data, double centerX, double centerY)
        {
            double x, y;
            if (data.RotateIsLeft)
            {
                x = centerX;
            }
            else
            {
                x = -centerX;
            }

            if (data.RotateIsTop)
            {
                y = centerY;
            }
            else
            {
                y = -centerY;
            }

            return Math.Atan2(x, y) - PI;
        }

        private double RadFromDeg(double angle)
        {
            return angle * PI / 180;
        }

        private double DegFromRad(double angle)
        {
            return angle * 180 / PI;
        }

        private void thumb_MouseMove(object sender, MouseEventArgs e)
        {
            DragData dragData;
            _mousePosRelativeToThis = e.GetPosition(this);
            if (!_lastMousePosInitialized)
            {
                _lastMousePos.X = _mousePosRelativeToThis.X;
                _lastMousePos.Y = _mousePosRelativeToThis.Y;
                _lastMousePosInitialized = true;
                return;
            }
            Point delta = new Point();
            delta.X = _mousePosRelativeToThis.X - _lastMousePos.X;
            delta.Y = _mousePosRelativeToThis.Y - _lastMousePos.Y;
            UpdateCursorPos(_mousePosRelativeToThis);
            switch (_currentDragOperation)
            {
                case DragType.Move:
                    _x += delta.X;
                    _y += delta.Y;
                    trTranslate.X = _x;
                    trTranslate.Y = _y;
                    if (_adornedObject != null)
                    {
                        _adornedObject.X = _x;
                        _adornedObject.Y = _y;
                    }
                    break;

                case DragType.Rotate:
                    {
                        double w2 = Width / 2;
                        double h2 = Height / 2;
                        Point centerOfRotation = new Point(_x + w2, _y + h2);
                        dragData = _dragData[_draggedThumb];
                        // calculate the angle between the mouse and the center of rotation and subtract the adorner angle to make sure
                        // the adorner, mouse and center of rotation will end up on the same line
                        _angle = DegFromRad(Math.Atan2(centerOfRotation.X - _mousePosRelativeToThis.X, -centerOfRotation.Y + _mousePosRelativeToThis.Y) +
                            dragData.RotateAdornerAngleRad);
                        trRotate.Angle = _angle;
                        if (_adornedObject != null)
                        {
                            _adornedObject.Angle = _angle;
                        }
                        UpdateCursorAngle();
                    }
                    break;

                case DragType.Size:
                    // resize algorithm: 
                    // 1. find the distance between the line passing through the point the mouse was last at 
                    //   and the current point - this is the distance to move the edge
                    // 2. Move the edge based on the current adorner (dragData)
                    Point rotatedDelta = trRotate.Inverse.Transform(delta);
                    dragData = _dragData[_draggedThumb];

                    double widthInc = rotatedDelta.X;
                    double heightInc = rotatedDelta.Y;

                    if (_preserveCenterOnResize)
                    {
                        // since we preserve the center, 
                        // width and height change should be double when the mouse if moved one pixel 
                        // so that the cursor will follow the mouse
                        widthInc *= dragData.WidthChange * 2;
                        heightInc *= dragData.HeightChange * 2;

                        // check for min width and height
                        if (Width + widthInc < 0)
                        {
                            widthInc = -Width;
                        }
                        if (Height + heightInc < 0)
                        {
                            heightInc = -Height;
                        }

                        double widthInc2 = widthInc / 2;
                        double heightInc2 = heightInc / 2;

                        Width += widthInc;
                        Height += heightInc;
                        if (_adornedObject != null)
                        {
                            _adornedObject.Width = Width;
                            _adornedObject.Height = Height;
                        }

                        // the difference between the center positions before the rotation is (-widthInc2, -widthInc2)
                        // the center is moved by (-widthInc2, -widthInc2) in order to compensate for the difference,
                        // which effectively will preserve the center
                        _x -= widthInc2;
                        _y -= heightInc2;
                    }
                    else
                    {
                        widthInc *= dragData.WidthChange;
                        heightInc *= dragData.HeightChange;

                        // check for min width and height
                        if (Width + widthInc < 0)
                        {
                            widthInc = -Width;
                        }
                        if (Height + heightInc < 0)
                        {
                            heightInc = -Height;
                        }

                        double widthInc2 = widthInc / 2;
                        double heightInc2 = heightInc / 2;

                        // HSign, VSign combinations:
                        // +, - = top-left
                        // -, - = top-right
                        // +, + = bototm-left
                        // -, + = bottom-right

                        Point vectorToNewCenterAfterRotation = new Point(dragData.HSignInverse * widthInc2, dragData.VSign * heightInc2);
                        Point pointToKeepOrigAfterSize = trRotate.Inverse.Transform(vectorToNewCenterAfterRotation);

                        double dx = pointToKeepOrigAfterSize.X - widthInc2;
                        double dy = pointToKeepOrigAfterSize.Y + heightInc2;

                        Width += widthInc;
                        Height += heightInc;
                        if (_adornedObject != null)
                        {
                            _adornedObject.Width = Width;
                            _adornedObject.Height = Height;
                        }

                        _x += dx;
                        _y -= dy;
                    }

                    trTranslate.X = _x;
                    trTranslate.Y = _y;
                    if (_adornedObject != null)
                    {
                        _adornedObject.X = _x;
                        _adornedObject.Y = _y;
                    }
                    break;

            }
            _lastMousePos.X = _mousePosRelativeToThis.X;
            _lastMousePos.Y = _mousePosRelativeToThis.Y;
        }

        bool _preserveCenterOnResize = false;

        public void StartMoveDrag()
        {
            _lastMousePosInitialized = false;
            thumb_DragStarted(thumbMove, null);
        }

        private void thumb_DragStarted(object sender, System.Windows.Controls.Primitives.DragStartedEventArgs e)
        {
            _currentDragOperation = _dragData[(Thumb)sender].DragType;
            _draggedThumb = (Thumb)sender;
        }

        private void thumb_DragCompleted(object sender, System.Windows.Controls.Primitives.DragCompletedEventArgs e)
        {
            _currentDragOperation = DragType.None;
            _draggedThumb = null;
        }

        private void thumbMove_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            _currentDragOperation = DragType.None;
            _draggedThumb = null;
        }

        private void HideAllCursors()
        {
            pathSize.Visibility = Visibility.Collapsed;
            pathRotate.Visibility = Visibility.Collapsed;
            pathCrosshair.Visibility = Visibility.Collapsed;
            _currentCursorData.Translate = null;
        }

        private void ShowCursor(DragType type, Point pos, double angle)
        {
            HideAllCursors();
            _currentCursorData.InitAngle = angle;
            switch (type)
            {
                case DragType.Rotate:
                    pathRotate.Visibility = Visibility.Visible;
                    _currentCursorData.Rotate = pathRotateRotate;
                    _currentCursorData.Translate = pathRotateTranslate;
                    break;

                case DragType.Size:
                    pathSize.Visibility = Visibility.Visible;
                    _currentCursorData.Rotate = pathSizeRotate;
                    _currentCursorData.Translate = pathSizeTranslate;
                    break;

                case DragType.Connector:
                    pathCrosshair.Visibility = Visibility.Visible;
                    _currentCursorData.Translate = pathCrosshairTranslate;
                    break;

                default:
                    throw new ArgumentException("Unsupported mouse cursor: " + type.ToString());
            }
            UpdateCursorAngle();
            UpdateCursorPos(pos);
        }

        private void UpdateCursorAngle()
        {
            _currentCursorData.Rotate.Angle = _currentCursorData.InitAngle + _angle;
        }

        private void UpdateCursorPos(Point pos)
        {
            UpdateCursorPos(pos.X, pos.Y);
        }

        private void UpdateCursorPos(double x, double y)
        {
            if (_currentCursorData.Translate == null)
            {
                return;
            }

            _currentCursorData.Translate.X = x;
            _currentCursorData.Translate.Y = y;
        }

        Thumb _draggedThumb = null;
        CursorData _currentCursorData = new CursorData();

        #endregion

        /// <summary>
        /// Clipping rectangle specifying the visible area of the glass border
        /// </summary>
        Rect _borderClipRect = new Rect(6, 6, 0, 0);
        private void gridControl_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            _borderClipRect.Width = e.NewSize.Width;
            _borderClipRect.Height = e.NewSize.Height;
            borderClip.Rect = _borderClipRect;
        }


        #region CursorData

        private class CursorData
        {
            public TranslateTransform Translate = null;
            public RotateTransform Rotate = null;
            public double InitAngle;
        };

        #endregion

    };
}
