﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.ComponentModel;
using System.Collections.Generic;

namespace DiagramLibrary.Diagram
{

    [TemplateVisualState(GroupName = "HoverStates", Name = "NoHover")]
    [TemplateVisualState(GroupName = "HoverStates", Name = "Hover")]
    [TemplateVisualState(GroupName = "SelectionStates", Name = "Unselected")]
    [TemplateVisualState(GroupName = "SelectionStates", Name = "Selected")]
    public class Connector : Control, INotifyPropertyChanged
    {
        // drag start point, relative to the DesignerCanvas
        private Point? dragStartPoint = null;

        // center position of this Connector relative to the DesignerCanvas
        private Point _position;
        public Point Position
        {
            get { return _position; }
            set
            {
                if (_position != value)
                {
                    _position = value;
                    OnPropertyChanged("Position");
                }
            }
        }


        #region Dependency Properties
        
        #region Is Selected

        public static readonly DependencyProperty IsSelectedProperty = DependencyProperty.Register(
            "IsSelected", typeof(bool), typeof(Connector), new PropertyMetadata(OnIsSelectedChanged));
        public bool IsSelected
        {
            get { return (bool)GetValue(IsSelectedProperty); }
            set { SetValue(IsSelectedProperty, value); }
        }
        private static void OnIsSelectedChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var control = d as Connector;
            control.UpdateVisualStates();
        }

        #endregion

        #region Is MouseOver

        public static readonly DependencyProperty IsMouseOverProperty = DependencyProperty.Register(
            "IsMouseOver", typeof(bool), typeof(Connector), new PropertyMetadata(OnMouseOver));
        public bool IsMouseOver
        {
            get { return (bool)GetValue(IsMouseOverProperty); }
            set { SetValue(IsMouseOverProperty, value); }
        }
        private static void OnMouseOver(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var control = d as Connector;
            control.UpdateVisualStates();
        }

        #endregion

        #endregion

        #region ctor

        public Connector()
        {
            DefaultStyleKey = typeof(Connector);
            // fired when layout changes
            base.LayoutUpdated += new EventHandler(Connector_LayoutUpdated);
        }

        // when the layout changes we update the position property
        void Connector_LayoutUpdated(object sender, EventArgs e)
        {
            DiagramPanel designer = GetDiagramPanel(this);
            if (designer != null)
            {
                //get centre position of this Connector relative to the DesignerCanvas
                this.Position = this.TransformToVisual(designer).Transform(new Point(this.Width / 2, this.Height / 2));
            }
        }

        #endregion

        //// the DesignerItem this Connector belongs to;
        //// retrieved from DataContext, which is set in the
        //// DiagramItem template
        //private DiagramItem _parentDiagramItem;
        //public DiagramItem ParentDiagramItem 
        //{
        //    get
        //    {
        //        if (_parentDiagramItem == null)
        //            _parentDiagramItem = this.DataContext as DiagramItem;

        //        return _parentDiagramItem;
        //    }
        //}

        internal DiagramItem ParentDiagramItem { get; set; }
        internal DiagramPanel Owner { get; set; }

        private Canvas ParentCanvas
        {
            get
            {
                if (this.Parent != null &&
                    this.Parent is Canvas)
                {
                    return this.Parent as Canvas;
                }
                else
                {
                    return null;
                }
            }
        }

        public ConnectorOrientation Orientation { get; set; }

        // keep track of connections that link to this connector
        private List<Connection> _connections;
        public List<Connection> Connections
        {
            get
            {
                if (_connections == null)
                {
                    _connections = new List<Connection>();
                }
                return _connections;
            }
        }


        protected override void OnMouseEnter(MouseEventArgs e)
        {
            IsMouseOver = true;
            UpdateVisualStates();
        }
        protected override void OnMouseLeave(MouseEventArgs e)
        {
            IsMouseOver = false;
            UpdateVisualStates();
        }

        private void UpdateVisualStates()
        {
            if (IsMouseOver)
            {
                VisualStateManager.GoToState(this, "Hover", true);
            }
            else
            {
                VisualStateManager.GoToState(this, "NoHover", true);
            }

            if (IsSelected)
            {
                VisualStateManager.GoToState(this, "Selected", true);
            }
            else
            {
                VisualStateManager.GoToState(this, "Unselected", true);
            }
        }

        internal ConnectorInfo GetInfo()
        {
            double offset = 5.0d;
            ConnectorInfo info = new ConnectorInfo()
            {
                DiagramItemLeft = Canvas.GetLeft(this.ParentDiagramItem) - offset,
                DiagramItemTop = Canvas.GetTop(this.ParentDiagramItem) - offset,
                Position = this.Position,
                //Position = this.TransformToVisual(this.Owner).Transform(new Point((this.Width - offset) / 2, (this.Height - offset) / 2)),
                Orientation = this.Orientation
            };
            
            if (this.ParentDiagramItem.ActualHeight > offset && this.ParentDiagramItem.ActualWidth > offset)
            {
                info.DiagramItemSize = new Size(this.ParentDiagramItem.ActualWidth - offset, this.ParentDiagramItem.ActualHeight - offset);
            }
            else
            {
                info.DiagramItemSize = new Size(this.ParentDiagramItem.ActualWidth, this.ParentDiagramItem.ActualHeight);
            }
            return info;
        }

        // iterate through visual tree to get parent DesignerCanvas
        private DiagramPanel GetDiagramPanel(DependencyObject element)
        {
            while (element != null && !(element is DiagramPanel))
                element = VisualTreeHelper.GetParent(element);

            return element as DiagramPanel;
        }

        #region INotifyPropertyChanged
        public event PropertyChangedEventHandler PropertyChanged;
        public void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }
        #endregion

    };

    /// <summary>
    /// Provides compact info about a connector; used for the
    /// routing algorithm, instead of passing the full 
    /// connector object.
    /// </summary>
    internal struct ConnectorInfo
    {
        public double DiagramItemLeft { get; set; }
        public double DiagramItemTop { get; set; }
        public Size DiagramItemSize { get; set; }
        public Point Position { get; set; }
        public ConnectorOrientation Orientation { get; set; }
    }
}
