﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;
using System.Collections;
using System.Collections.Specialized;
using System.Windows.Data;
using System.Linq;
using System.Collections.ObjectModel;

namespace DiagramLibrary.Diagram
{
    [TemplateVisualState(GroupName = "HoverStates", Name = "NoHover")]
    [TemplateVisualState(GroupName = "HoverStates", Name = "Hover")]
    [TemplateVisualState(GroupName = "SelectionStates", Name = "Unselected")]
    [TemplateVisualState(GroupName = "SelectionStates", Name = "Selected")]
    public class DiagramItem : ContentControl
    {
        #region Member Variables

        private bool _isDragging;
		private Point _startPosition;

        #endregion

        #region Dependency Properties

        #region Is Selected

        public static readonly DependencyProperty IsSelectedProperty = DependencyProperty.Register(
            "IsSelected", typeof(bool), typeof(DiagramItem), new PropertyMetadata(OnIsSelectedChanged));
        public bool IsSelected
        {
            get { return (bool)GetValue(IsSelectedProperty); }
            set { SetValue(IsSelectedProperty, value); }
        }
        private static void OnIsSelectedChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var control = d as DiagramItem;
            control.UpdateVisualStates();
            foreach (Connector cnt in control.Connectors)
            {
                cnt.IsMouseOver = control.IsSelected;
            }
        }
        #endregion

        #region Connection

        private ObservableCollection<Connection> _connections = new ObservableCollection<Connection>();
        public ObservableCollection<Connection> Connections
        {
            get { return _connections; }
        }

        #endregion

        #region Connectors

        private ObservableCollection<Connector> _connectors = new ObservableCollection<Connector>();
        public ObservableCollection<Connector> Connectors
        {
            get { return _connectors; }
        }

        #endregion

        #endregion

        internal DiagramItem ParentItem { get; set; }
		internal DiagramPanel Owner { get; set; }
		internal bool IsConnectorInitialized { get; set; }
		public bool IsMouseOver { get; set; }
        public int Id { get; set; }
        public int ParentId { get; set; }
        public double Top { get; set; }
        public double Left { get; set; }

        private ObservableCollection<DiagramItem> _childItems = new ObservableCollection<DiagramItem>();
        public ObservableCollection<DiagramItem> ChildItems 
        {
            get { return _childItems; }
        }


        #region ctor

        public DiagramItem()
        {
            DefaultStyleKey = typeof(DiagramItem);
        }

        #endregion

		protected override void OnMouseEnter(MouseEventArgs e)
		{
			IsMouseOver = true;
            foreach (Connector cnt in Connectors)
            {
                cnt.IsMouseOver = true;
            }
            UpdateVisualStates();
		}
		protected override void OnMouseLeave(MouseEventArgs e)
		{
			IsMouseOver = false;
            foreach (Connector cnt in Connectors)
            {
                if (this.IsSelected == false)
                {
                    cnt.IsMouseOver = false;    
                }
            }
            UpdateVisualStates();
		}
		protected override void OnMouseMove(MouseEventArgs e)
		{
			if (_isDragging)
			{
				var currentPosition = e.GetPosition(ParentItem);
				var delta = new Point(currentPosition.X - _startPosition.X, currentPosition.Y - _startPosition.Y);
				_startPosition = currentPosition;

				var moveGroup = (Keyboard.Modifiers & ModifierKeys.Control) == ModifierKeys.Control;
				MoveItemBy(delta, moveGroup);
			}
		}
		protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
		{
			_isDragging = true;

			CaptureMouse();
			_startPosition = e.GetPosition(ParentItem);
			Owner.SetSelection(this);

			UpdateVisualStates();
		}
		protected override void OnMouseLeftButtonUp(MouseButtonEventArgs e)
		{
			_isDragging = false;
			ReleaseMouseCapture();
		}

		private void UpdateVisualStates()
		{
			if (IsMouseOver)
			{
				VisualStateManager.GoToState(this, "Hover", true);
			}
			else
			{
				VisualStateManager.GoToState(this, "NoHover", true);
			}

			if (IsSelected)
			{
				VisualStateManager.GoToState(this, "Selected", true);
			}
			else
			{
				VisualStateManager.GoToState(this, "Unselected", true);
			}

            //
            //  Added the following to set the zIndex.
            //
            Canvas.SetZIndex(this, Owner.GetMaxZIndex + 1);
		}

		public void MoveItemBy(Point delta, bool moveGroup)
		{
			var x = Canvas.GetLeft(this) + delta.X;
			var y = Canvas.GetTop(this) + delta.Y;

            Canvas.SetLeft(this, x);
            Canvas.SetTop(this, y);

            //UpdateConnection();
            UpdateConnectors();
            //UpdateConnection();
        }

		private void MoveItemTo(Point point)
		{
			var currX = Canvas.GetLeft(this);
			var currY = Canvas.GetTop(this);
			var delta = new Point(point.X - currX, point.Y - currY);

			MoveItemBy(delta, false);
		}

		private void UpdateConnection()
		{
            if (ParentItem != null)
            {
                foreach (Connection item in Connections)
                {
                    var cnt = item;
                    UpdateConnectionPath(cnt);
                }
            }

            foreach (var childItem in ChildItems)
            {
                var current = childItem;

                foreach (Connection item in current.Connections)
                {
                    var cnt = item;
                    UpdateConnectionPath(cnt);
                }
            }
        }

        public void UpdateConnectionPath(Connection cnt)
        {
            cnt.UpdatePathGeometry();


            //cnt.Path.X1 = Canvas.GetLeft(cnt.Source) + cnt.Source.ActualWidth / 2;
            //cnt.Path.Y1 = Canvas.GetTop(cnt.Source) + 5;

            //cnt.Path.X2 = Canvas.GetLeft(cnt.Sink) + cnt.Sink.ActualWidth / 2;
            //cnt.Path.Y2 = Canvas.GetTop(cnt.Sink) + 5;
        }

        private void UpdateConnectors()
        {
            foreach (Connector item in Connectors)
            {
                var cnt = item;
                double left;
                double top;
                switch (cnt.Orientation)
                {
                    case ConnectorOrientation.Left:
                        break;
                    case ConnectorOrientation.Top:
                        left = Canvas.GetLeft(this) + (ActualWidth / 2) - (cnt.Width / 2);
                        top = Canvas.GetTop(this) - 10;
                        Canvas.SetLeft(cnt, left);
                        Canvas.SetTop(cnt, top);
                        cnt.Position = new Point(left, top);
                        break;
                    case ConnectorOrientation.Right:
                        break;
                    case ConnectorOrientation.Bottom:
                        left = Canvas.GetLeft(this) + (ActualWidth / 2) - (cnt.Width / 2);
                        top = Canvas.GetTop(this) + ActualHeight;
                        Canvas.SetLeft(cnt, left);
                        Canvas.SetTop(cnt, top);
                        cnt.Position = new Point(left, top);
                        break;
                }
            }
        }
    };
}
