﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;

namespace DiagramLibrary.Diagram
{
    /// <summary>
    /// This code came from the following blog post:
    /// http://www.charlespetzold.com/blog/2008/11/Text-on-a-Path-in-Silverlight.html
    /// </summary>
    public class PathFigureHelper
    {
        // Each segment is part of a flattened PathFigure. The offset is the length
        //  from the beginning.
        struct Segment
        {
            public readonly double Offset;
            public readonly Point Point;

            public Segment(double offset, Point point)
            {
                Offset = offset;
                Point = point;
            }
        };

        class SegmentList : List<Segment>
        {
            public void Add(Point point)
            {
                if (Count == 0)
                    Add(new Segment(0, point));
                else
                {
                    // Determine accumulated offset of Point
                    Segment segmentPrevious = this[Count - 1];
                    double length = GetLineLength(point, segmentPrevious.Point);
                    Add(new Segment(segmentPrevious.Offset + length, point));
                }
            }
        }

        SegmentList segments = new SegmentList();

        // Does flattening and fills 'segments' collections
        public void SetPathFigure(PathFigure pathFigure)
        {
            segments.Clear();
            segments.Add(pathFigure.StartPoint);
            Point ptPrevious = pathFigure.StartPoint;

            foreach (PathSegment pathSegment in pathFigure.Segments)
            {
                if (pathSegment is LineSegment)
                {
                    LineSegment lineSegment = pathSegment as LineSegment;
                    segments.Add(lineSegment.Point);
                    ptPrevious = lineSegment.Point;
                }
                else if (pathSegment is PolyLineSegment)
                {
                    PointCollection points = (pathSegment as PolyLineSegment).Points;
                    foreach (Point point in points)
                    {
                        segments.Add(point);
                        ptPrevious = point;
                    }
                }
                else if (pathSegment is QuadraticBezierSegment)
                {
                    QuadraticBezierSegment quadraticBezierSegment = pathSegment as QuadraticBezierSegment;
                    FlattenQuadraticBezier(ptPrevious, quadraticBezierSegment.Point1, quadraticBezierSegment.Point2);
                    ptPrevious = quadraticBezierSegment.Point2;
                }
                else if (pathSegment is PolyQuadraticBezierSegment)
                {
                    PointCollection points = (pathSegment as PolyQuadraticBezierSegment).Points;
                    for (int i = 0; i < points.Count; i += 2)
                    {
                        FlattenQuadraticBezier(ptPrevious, points[i], points[i + 1]);
                        ptPrevious = points[i + 1];
                    }
                }
                else if (pathSegment is BezierSegment)
                {
                    BezierSegment bezierSegment = pathSegment as BezierSegment;
                    FlattenBezier(ptPrevious, bezierSegment.Point1, bezierSegment.Point2, bezierSegment.Point3);
                    ptPrevious = bezierSegment.Point3;
                }
                else if (pathSegment is PolyBezierSegment)
                {
                    PointCollection points = (pathSegment as PolyBezierSegment).Points;

                    for (int i = 0; i < points.Count; i += 3)
                    {
                        FlattenBezier(ptPrevious, points[i], points[i + 1], points[i + 2]);
                        ptPrevious = points[i + 2];
                    }
                }
                else
                {
                    throw new NotImplementedException(
                        String.Format("PathSegment == {0} has not been implemented",
                                      pathSegment.GetType().Name));
                }
            }
        }

        void FlattenQuadraticBezier(Point pt0, Point pt1, Point pt2)
        {
            int num = (int)(GetLineLength(pt0, pt1) + GetLineLength(pt1, pt2));

            for (int i = 1; i < num; i++)
            {
                double t = i / (num - 1.0);

                double x = (1 - t) * (1 - t) * pt0.X +
                           2 * t * (1 - t) * pt1.X +
                           t * t * pt2.X;

                double y = (1 - t) * (1 - t) * pt0.Y +
                           2 * t * (1 - t) * pt1.Y +
                           t * t * pt2.Y;

                segments.Add(new Point(x, y));
            }
        }

        void FlattenBezier(Point pt0, Point pt1, Point pt2, Point pt3)
        {
            int num = (int)(GetLineLength(pt0, pt1) + GetLineLength(pt1, pt2) + GetLineLength(pt2, pt3));

            for (int i = 1; i < num; i++)
            {
                double t = i / (num - 1.0);

                double x = (1 - t) * (1 - t) * (1 - t) * pt0.X +
                           3 * t * (1 - t) * (1 - t) * pt1.X +
                           3 * t * t * (1 - t) * pt2.X +
                           t * t * t * pt3.X;

                double y = (1 - t) * (1 - t) * (1 - t) * pt0.Y +
                           3 * t * (1 - t) * (1 - t) * pt1.Y +
                           3 * t * t * (1 - t) * pt2.Y +
                           t * t * t * pt3.Y;

                segments.Add(new Point(x, y));
            }
        }

        public static double GetLineLength(Point pt1, Point pt2)
        {
            double diffX = pt2.X - pt1.X;
            double diffY = pt2.Y - pt1.Y;
            return Math.Sqrt(diffX * diffX + diffY * diffY);
        }

        public double Length
        {
            get
            {
                if (segments.Count < 2)
                    return 0;

                return segments[segments.Count - 1].Offset;
            }
        }

        public void GetPointAtFractionLength(double progress, out Point point, out Point tangent)
        {
            if (progress < 0 || progress > 1 || Length == 0)
            {
                point = new Point();
                tangent = new Point();
                return;
            }

            double offset = progress * Length;

            // Binary search
            int index1 = 0;
            int index2 = segments.Count - 1;

            while (index2 - index1 > 1)
            {
                int testIndex = (index2 + index1) / 2;

                if (offset < segments[testIndex].Offset)
                    index2 = testIndex;
                else
                    index1 = testIndex;
            }
            Point point1 = segments[index1].Point;
            Point point2 = segments[index2].Point;
            double offset1 = segments[index1].Offset;
            double offset2 = segments[index2].Offset;

            point = new Point(((offset2 - offset) * point1.X +
                               (offset - offset1) * point2.X) / (offset2 - offset1),

                              ((offset2 - offset) * point1.Y +
                               (offset - offset1) * point2.Y) / (offset2 - offset1));

            tangent = new Point(point2.X - point1.X, point2.Y - point1.Y);
        }
    };
}
