﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace System.Windows
{
    public static class Extensions
    {
        #region Rect

        public static Point TopLeft(this Rect value)
        {
            return new Point(value.Left, value.Top);
        }
        public static Point TopRight(this Rect value)
        {
            return new Point(value.Left + value.Width, value.Top);
        }
        public static Point BottomLeft(this Rect value)
        {
            return new Point(value.Left, value.Top + value.Height);
        }
        public static Point BottomRight(this Rect value)
        {
            return new Point(value.Left + value.Width, value.Top + value.Height);
        }

        /// <summary>
        /// Indicates whether the specified rectangle intersects with the current rectangle.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="rect">The rectangle to check.</param>
        /// <returns>true if the specified rectangle intersects with the current rectangle; otherwise, false.</returns>
        /// <remarks>
        /// This method is not available for Silverlight and was taken from
        /// both the WPF and Silverlight methods.
        /// Silverlight:  http://msdn.microsoft.com/en-us/library/ms558125(v=VS.95).aspx
        /// WPF: http://msdn.microsoft.com/en-us/library/system.windows.rect.intersectswith.aspx
        /// </remarks>
        public static bool IntersectsWith(this Rect value, Rect rect)
        {
            // Silverlght does not have this method so we are creating it according to the documentation from
            // both the WPF and Silverlight methods.
            // Silverlight:  http://msdn.microsoft.com/en-us/library/ms558125(v=VS.95).aspx
            // WPF: http://msdn.microsoft.com/en-us/library/system.windows.rect.intersectswith.aspx
            Rect currentRect = new Rect(value.Left, value.Top, value.Width, value.Height);
            currentRect.Intersect(rect);
            return !currentRect.IsEmpty;
        }
        /// <summary>
        /// Expands or shrinks the rectangle by using the specified width and height
        /// amounts, in all directions.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="width">The amount by which to expand or shrink the left and right sides of the rectangle.</param>
        /// <param name="height">The amount by which to expand or shrink the top and bottom sides of the rectangle.</param>
        /// <remarks>
        /// This method is not available for Silverlight and was taken from the 
        /// WPF version.  It tries to follow the WPF implementation as closely
        /// as possible.
        /// http://msdn.microsoft.com/en-us/library/ms558114.aspx
        /// The Width of the resulting rectangle is increased or decreased by twice the specified width offset, because 
        /// it is applied to both the left and right sides of the rectangle. Likewise, the Height of the resulting 
        /// rectangle is increased or decreased by twice the specified height.
        /// 
        /// If the specified width or height shrink the rectangle by more than its current Width or Height—giving 
        /// the rectangle a negative area—the rectangle becomes the Empty rectangle.
        /// </remarks>
        public static Rect Inflate(this Rect value, double width, double height)
        {
            if (value.IsEmpty) throw new InvalidOperationException("Rect cannot be empty.");

            double fullWidth = width * 2;
            double fullHeight = height * 2;

            if (value.Width + fullWidth < 0) fullWidth = 0d;
            if (value.Height + fullHeight < 0) fullHeight = 0d;
            if (fullWidth == 0d && fullHeight == 0d) value = Rect.Empty;

            value.Width += fullWidth;
            value.Height += fullHeight;

            return value;
        }

        #endregion

        #region Point

        /// <summary>
        /// Offsets a point's X and Y coordinates by the specified amounts.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="offsetX">The amount to offset the point's
        /// X coordinate. </param>
        /// <param name="offsetY">The amount to offset the point's
        /// Y coordinate.</param>
        /// <remarks>
        /// This operation is equivalent to adding a Point to a Vector.
        /// But in Silverlight we don't have a Vector object.
        /// This method is not available for Silverlight and was taken from
        /// the WPF method:
        /// http://msdn.microsoft.com/en-us/library/system.windows.point.offset.aspx
        /// </remarks>
        public static void Offset(this Point value, double offsetX, double offsetY)
        {
            value.X += offsetX;
            value.Y += offsetY;
        }
        #endregion

    };

    public static class PointHelper
    {
        /// <summary>
        /// This method simulates the Point.Subtract(point1, point2).Length
        /// property.  It takes two points and returns the Length after performing
        /// a subtraction.
        /// http://msdn.microsoft.com/en-us/library/ms557949.aspx
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <returns></returns>
        public static double SubtractAndReturnLength(Point point1, Point point2)
        {
            Point pt = new Point(point1.X - point2.X, point1.Y - point2.Y);
            // Get the length of the vector.  Point does not have
            // the Length property, we will calculate it.
            // e.g. Vector(20, 30) --> Length = 36.0555
            Double result = Math.Sqrt((pt.X * pt.X) + (pt.Y * pt.Y));
            return result;

        }

    };
}
namespace System
{
    public static class Extensions
    {
        #region Object

        public static double Left(this Object value)
        {
            var g = ((FrameworkElement)value).TransformToVisual(Application.Current.RootVisual);
            var offset = g.Transform(new Point(0, 0));
            return offset.X;
        }
        public static double Left(this Object value, UIElement parent)
        {
            var g = ((FrameworkElement)value).TransformToVisual(parent);
            var offset = g.Transform(new Point(0, 0));
            return offset.X;
        }

        public static double Top(this Object value)
        {
            var g = ((FrameworkElement)value).TransformToVisual(Application.Current.RootVisual);
            var offset = g.Transform(new Point(0, 0));
            return offset.Y;
        }
        public static double Top(this Object value, UIElement parent)
        {
            var g = ((FrameworkElement)value).TransformToVisual(parent);
            var offset = g.Transform(new Point(0, 0));
            return offset.Y;
        }

        #endregion
    };
}
